@@ -14,17 +14,15 @@ use Moo;
 
 use IPC::Run3; # For run3().
 
-use Set::Array;
-
 use Try::Tiny;
 
-use Types::Standard qw/Any HashRef Int Str/;
+use Types::Standard qw/Any ArrayRef HashRef Int Str/;
 
 has command =>
 (
-	default  => sub{return Set::Array -> new},
-	is       => 'rw',
-	isa      => Any,
+	default  => sub{[]},
+	is       => 'ro',
+	isa      => ArrayRef,
 	required => 0,
 );
 
@@ -110,9 +108,9 @@ has node_hash =>
 
 has scope =>
 (
-	default  => sub{return Set::Array -> new},
-	is       => 'rw',
-	isa      => Any,
+	default  => sub{[]},
+	is       => 'ro',
+	isa      => ArrayRef,
 	required => 0,
 );
 
@@ -140,7 +138,7 @@ has valid_attributes =>
 	required => 0,
 );
 
-our $VERSION = '2.50';
+our $VERSION = '2.51';
 
 # -----------------------------------------------
 
@@ -151,39 +149,38 @@ sub BUILD
 	my($dot)     = which('dot');
 	my($global)  =
 	{
-		directed		=> $$globals{directed}			? 'digraph'				: 'graph',
-		driver			=> $$globals{driver}			? $$globals{driver}		: $dot,
-		format			=> $$globals{format}			? $$globals{format}		: 'svg',
-		im_format		=> $$globals{im_format}			? $$globals{im_format}	: 'cmapx',
-		label			=> $$globals{directed}			? '->'					: '--',
-		name			=> defined($$globals{name})		? $$globals{name}		: 'Perl',
+		directed		=> $$globals{directed} ? 'digraph' : 'graph',
+		driver			=> $$globals{driver} || $dot,
+		format			=> $$globals{format} ||	'svg',
+		im_format		=> $$globals{im_format} || 'cmapx',
+		label			=> $$globals{directed} ? '->' : '--',
+		name			=> $$globals{name} // 'Perl',
 		record_shape	=> ($$globals{record_shape} && $$globals{record_shape} =~ /^(M?record)$/) ? $1 : 'Mrecord',
-		strict			=> defined($$globals{strict})	? $$globals{strict}		:  0,
-		subgraph		=> $$globals{subgraph}			? $$globals{subgraph}	: {},
-		timeout			=> defined($$globals{timeout})	? $$globals{timeout}	: 10,
+		strict			=> $$globals{strict} //  0,
+		subgraph		=> $$globals{subgraph} || {},
+		timeout			=> $$globals{timeout} // 10,
 	};
 	my($im_metas)	= $self -> im_meta;
 	my($im_meta)	=
 	{
-		URL => $$im_metas{URL} ? $$im_metas{URL} : '',
+		URL => $$im_metas{URL} || '',
 	};
 
 	$self -> global($global);
 	$self -> im_meta($im_meta);
 	$self -> load_valid_attributes;
-	$self -> validate_params('global',		%{$self -> global});
-	$self -> validate_params('graph',		%{$self -> graph});
-	$self -> validate_params('im_meta',		%{$self -> im_meta});
-	$self -> validate_params('node',		%{$self -> node});
-	$self -> validate_params('edge',		%{$self -> edge});
-	$self -> validate_params('subgraph',	%{$self -> subgraph});
-	$self -> scope -> push
-		({
-			edge     => $self -> edge,
-			graph    => $self -> graph,
-			node     => $self -> node,
-			subgraph => $self -> subgraph,
-		 });
+	$self->validate_params('global',	$self->global);
+	$self->validate_params('graph',		$self->graph);
+	$self->validate_params('im_meta',	$self->im_meta);
+	$self->validate_params('node',		$self->node);
+	$self->validate_params('edge',		$self->edge);
+	$self->validate_params('subgraph',	$self->subgraph);
+	push @{ $self->scope }, {
+		edge     => $self -> edge,
+		graph    => $self -> graph,
+		node     => $self -> node,
+		subgraph => $self -> subgraph,
+	 };
 
 	my(%global)		= %{$self -> global};
 	my(%im_meta)	= %{$self -> im_meta};
@@ -194,14 +191,14 @@ sub BUILD
 	my($command) = (${$self -> global}{strict} ? 'strict ' : '')
 		. (${$self -> global}{directed} . ' ')
 		. ${$self -> global}{name}
-		. "\n{\n";
+		. " {\n";
 
 	for my $key (grep{$im_meta{$_} } sort keys %im_meta)
 	{
 		$command .= qq|$key = "$im_meta{$key}"; \n|;
 	}
 
-	$self -> command -> push($command);
+	push @{ $self->command }, $command;
 
 	$self -> default_graph;
 	$self -> default_node;
@@ -214,18 +211,14 @@ sub BUILD
 sub add_edge
 {
 	my($self, %arg) = @_;
-	my($from)   = delete $arg{from};
-	$from       = defined($from) ? $from : '';
-	my($to)     = delete $arg{to};
-	$to         = defined($to) ? $to : '';
-	my($label)  = defined($arg{label}) ? $arg{label} : '';
-	$label      =~ s/^\s+(<)/$1/;
-	$label      =~ s/(>)\s+$/$1/;
-	$label      =~ s/^(<)\n/$1/;
-	$label      =~ s/\n(>)$/$1/;
+	my $from    = delete $arg{from} // '';
+	my $to      = delete $arg{to} // '';
+	my $label   = $arg{label} // '';
+	$label      =~ s/^\s*(<)\n?/$1/;
+	$label      =~ s/\n?(>)\s*$/$1/;
 	$arg{label} = $label if (defined $arg{label});
 
-	$self -> validate_params('edge', %arg);
+	$self->validate_params('edge', \%arg);
 
 	# If either 'from' or 'to' is unknown, add a new node.
 
@@ -272,10 +265,6 @@ sub add_edge
 		}
 	}
 
-	# Add these nodes to the hashref of all nodes, if necessary.
-
-	$self -> node_hash($node) if ($new);
-
 	# Add this edge to the hashref of all edges.
 
 	my($edge)          = $self -> edge_hash;
@@ -289,35 +278,38 @@ sub add_edge
 		to_port    => $node[1][1],
 	};
 
-	$self -> edge_hash($edge);
-
 	# Add this edge to the DOT output string.
 
 	my($dot) = $self -> stringify_attributes(qq|"$from"$node[0][1] ${$self -> global}{label} "$to"$node[1][1]|, {%arg});
 
-	$self -> command -> push($dot);
+	push @{ $self->command }, _indent($dot, $self->scope);
 	$self -> log(debug => "Added edge: $dot");
 
 	return $self;
 
 } # End of add_edge.
 
+sub _indent {
+	my ($text, $scope) = @_;
+	return '' if $text !~ /\S/;
+	(' ' x @$scope) . $text;
+}
+
 # -----------------------------------------------
 
 sub add_node
 {
 	my($self, %arg) = @_;
-	my($name) = delete $arg{name};
-	$name     = defined($name) ? $name : '';
+	my $name = delete $arg{name} // '';
 
-	$self -> validate_params('node', %arg);
+	$self->validate_params('node', \%arg);
 
 	my($node)                 = $self -> node_hash;
-	$$node{$name}             = {} if (! $$node{$name});
-	$$node{$name}{attributes} = {} if (! $$node{$name}{attributes});
+	$$node{$name}             ||= {};
+	$$node{$name}{attributes} ||= {};
 	$$node{$name}{attributes} = {%{$$node{$name}{attributes} }, %arg};
 	%arg                      = %{$$node{$name}{attributes} };
-	my($label)                = defined($arg{label}) ? $arg{label} : '';
+	my($label)                = $arg{label} // '';
 	$label                    =~ s/^\s+(<)/$1/;
 	$label                    =~ s/(>)\s+$/$1/;
 	$label                    =~ s/^(<)\n/$1/;
@@ -377,8 +369,7 @@ sub add_node
 	$$node{$name}{attributes} = {%arg};
 	my($dot)                  = $self -> stringify_attributes(qq|"$name"|, {%arg});
 
-	$self -> command -> push($dot);
-	$self -> node_hash($node);
+	push @{ $self->command }, _indent($dot, $self->scope);
 	$self -> log(debug => "Added node: $dot");
 
 	return $self;
@@ -391,14 +382,12 @@ sub default_edge
 {
 	my($self, %arg) = @_;
 
-	$self -> validate_params('edge', %arg);
+	$self->validate_params('edge', \%arg);
 
-	my($scope)    = $self -> scope -> last;
+	my $scope    = $self->scope->[-1];
 	$$scope{edge} = {%{$$scope{edge} }, %arg};
-	my($tos)      = $self -> scope -> length - 1;
 
-	$self -> command -> push($self -> stringify_attributes('edge', $$scope{edge}) );
-	$self -> scope -> fill($scope, $tos, 1);
+	push @{ $self->command }, _indent($self->stringify_attributes('edge', $$scope{edge}), $self->scope);
 	$self -> log(debug => 'Default edge: ' . join(', ', map{"$_ => $$scope{edge}{$_}"} sort keys %{$$scope{edge} }) );
 
 	return $self;
@@ -411,14 +400,12 @@ sub default_graph
 {
 	my($self, %arg) = @_;
 
-	$self -> validate_params('graph', %arg);
+	$self->validate_params('graph', \%arg);
 
-	my($scope)     = $self -> scope -> last;
+	my $scope    = $self->scope->[-1];
 	$$scope{graph} = {%{$$scope{graph} }, %arg};
-	my($tos)       = $self -> scope -> length - 1;
 
-	$self -> command -> push($self -> stringify_attributes('graph', $$scope{graph}) );
-	$self -> scope -> fill($scope, $tos, 1);
+	push @{ $self->command }, _indent($self->stringify_attributes('graph', $$scope{graph}), $self->scope);
 	$self -> log(debug => 'Default graph: ' . join(', ', map{"$_ => $$scope{graph}{$_}"} sort keys %{$$scope{graph} }) );
 
 	return $self;
@@ -431,14 +418,12 @@ sub default_node
 {
 	my($self, %arg) = @_;
 
-	$self -> validate_params('node', %arg);
+	$self->validate_params('node', \%arg);
 
-	my($scope)    = $self -> scope -> last;
+	my $scope    = $self->scope->[-1];
 	$$scope{node} = {%{$$scope{node} }, %arg};
-	my($tos)      = $self -> scope -> length - 1;
 
-	$self -> command -> push($self -> stringify_attributes('node', $$scope{node}) );
-	$self -> scope -> fill($scope, $tos, 1);
+	push @{ $self->command }, _indent($self->stringify_attributes('node', $$scope{node}), $self->scope);
 	$self -> log(debug => 'Default node: ' . join(', ', map{"$_ => $$scope{node}{$_}"} sort keys %{$$scope{node} }) );
 
 	return $self;
@@ -451,14 +436,12 @@ sub default_subgraph
 {
 	my($self, %arg) = @_;
 
-	$self -> validate_params('subgraph', %arg);
+	$self->validate_params('subgraph', \%arg);
 
-	my($scope)        = $self -> scope -> last;
+	my $scope    = $self->scope->[-1];
 	$$scope{subgraph} = {%{$$scope{subgraph} }, %arg};
-	my($tos)          = $self -> scope -> length - 1;
 
-	$self -> command -> push($self -> stringify_attributes('subgraph', $$scope{subgraph}) );
-	$self -> scope -> fill($scope, $tos, 1);
+	push @{ $self->command }, _indent($self->stringify_attributes('subgraph', $$scope{subgraph}), $self->scope);
 	$self -> log(debug => 'Default subgraph: ' . join(', ', map{"$_ => $$scope{subgraph}{$_}"} sort keys %{$$scope{subgraph} }) );
 
 	return $self;
@@ -627,8 +610,8 @@ sub pop_subgraph
 {
 	my($self) = @_;
 
-	$self -> command -> push("}\n");
-	$self -> scope -> pop;
+	pop @{ $self->scope };
+	push @{ $self->command }, _indent("}\n", $self->scope);
 
 	return $self;
 
@@ -642,21 +625,21 @@ sub push_subgraph
 	my($name) = delete $arg{name};
 	$name     = defined($name) && length($name) ? qq|"$name"| : '';
 
-	$self -> validate_params('graph',    %{$arg{graph} });
-	$self -> validate_params('node',     %{$arg{node} });
-	$self -> validate_params('edge',     %{$arg{edge} });
-	$self -> validate_params('subgraph', %{$arg{subgraph} });
+	$self->validate_params('graph',    $arg{graph});
+	$self->validate_params('node',     $arg{node});
+	$self->validate_params('edge',     $arg{edge});
+	$self->validate_params('subgraph', $arg{subgraph});
 
 	# Child inherits parent attributes.
 
-	my($scope)        = $self -> scope -> last;
-	$$scope{edge}     = {%{$$scope{edge} },     %{$arg{edge} } };
-	$$scope{graph}    = {%{$$scope{graph} },    %{$arg{graph} } };
-	$$scope{node}     = {%{$$scope{node} },     %{$arg{node} } };
-	$$scope{subgraph} = {%{$$scope{subgraph} }, %{$arg{subgraph} } };
+	my $scope        = $self->scope->[-1];
+	$$scope{edge}     = {%{$$scope{edge} || {}}, %{$arg{edge} || {}}};
+	$$scope{graph}    = {%{$$scope{graph} || {}}, %{$arg{graph} || {}}};
+	$$scope{node}     = {%{$$scope{node} || {}}, %{$arg{node} || {}}};
+	$$scope{subgraph} = {%{$$scope{subgraph} || {}}, %{$arg{subgraph} || {}}};
 
-	$self -> scope -> push($scope);
-	$self -> command -> push(qq|\nsubgraph $name\n{\n|);
+	push @{ $self->command }, "\n" . _indent(join(' ', grep length, "subgraph", $name, "{\n"), $self->scope);
+	push @{ $self->scope }, $scope;
 	$self -> default_graph;
 	$self -> default_node;
 	$self -> default_edge;
@@ -764,14 +747,14 @@ sub run
 	$prefix				=~ s/:.+$//; # In case of 'png:gd', etc.
 	%arg				= ($prefix => 1);
 
-	$self -> validate_params('output_format', %arg);
+	$self->validate_params('output_format', \%arg);
 
 	my($prefix_1)	= $im_format;
 	$prefix_1		=~ s/:.+$//; # In case of 'png:gd', etc.
 	%arg			= ($prefix_1 => 1);
 
-	$self -> validate_params('output_format', %arg);
-	$self -> dot_input(join('', @{$self -> command -> print}) . "}\n");
+	$self->validate_params('output_format', \%arg);
+	$self -> dot_input(join('', @{ $self->command }) . "}\n");
 	$self -> log(debug => $self -> dot_input);
 
 	# Warning: Do not use $im_format in this 'if', because it has a default value.
@@ -889,7 +872,7 @@ sub stringify_attributes
 
 	for my $key (sort keys %$option)
 	{
-		$$option{$key} = '' if (! defined $$option{$key});
+		$$option{$key} //= '';
 		$$option{$key} =~ s/^\s+(<)/$1/;
 		$$option{$key} =~ s/(>)\s+$/$1/;
 		$dot           .= ($$option{$key} =~ /^<.+>$/s) ? qq|$key=$$option{$key} | : qq|$key="$$option{$key}" |;
@@ -916,12 +899,12 @@ sub stringify_attributes
 
 sub validate_params
 {
-	my($self, $context, %attributes) = @_;
-	my(%attr) = %{$self -> valid_attributes};
+	my($self, $context, $attributes) = @_;
+	my $valid = $self->valid_attributes;
 
-	for my $a (sort keys %attributes)
+	for my $a (sort keys %$attributes)
 	{
-		next if ($attr{$context}{$a} || ( ($context eq 'subgraph') && $attr{cluster}{$a}) );
+		next if ($valid->{$context}{$a} || ( ($context eq 'subgraph') && $valid->{cluster}{$a}) );
 
 		$self -> log(error => "Error: '$a' is not a valid attribute in the '$context' context");
 	}
@@ -944,11 +927,7 @@ GraphViz2 - A wrapper for AT&T's Graphviz
 
 =head2 Sample output
 
-Unpack the distro and copy html/*.html and html/*.svg to your web server's doc root directory.
-
-Then, point your browser at 127.0.0.1/index.html.
-
-Or, hit L<the demo page|http://savage.net.au/Perl-modules/html/graphviz2/index.html>.
+See L<https://graphviz-perl.github.io/>.
 
 =head2 Perl code
 
@@ -1876,7 +1855,7 @@ $context is one of 'edge', 'graph', 'node', or a special string. See the code fo
 
 You wouldn't normally need to use this method.
 
-=head2 validate_params($context, %attributes)
+=head2 validate_params($context, \%attributes)
 
 Validate the given attributes within the given context.
 
@@ -2516,10 +2495,6 @@ Outputs to ./html/utf8.2.svg by default.
 
 =over 4
 
-=item o Does GraphViz2 need to emulate the sort option in GraphViz?
-
-That depends on what that option really does.
-
 =item o Handle edges such as 1 -> 2 -> {A B}, as seen in L<Graphviz|http://www.graphviz.org/>'s graphs/directed/switch.gv
 
 But how?
